@@ -9,7 +9,6 @@
 
 [@@@warning "-27"]
 
-open CamomileLibraryDefault.Camomile
 open Lwt_react
 open LTerm_text
 open LTerm_geom
@@ -103,6 +102,15 @@ let default_keywords = [
 let keywords = ref (String_set.of_list default_keywords)
 let add_keyword kwd = keywords := String_set.add kwd !keywords
 
+(* +-----------------------------------------------------------------+
+   | Span of Lines                                                   |
+   +-----------------------------------------------------------------+ *)
+
+type lines = {
+  start: int;
+  stop:  int;
+}
+
 (* +-----------------------------------------------------------------+
    | Error reporting                                                 |
    +-----------------------------------------------------------------+ *)
@@ -126,9 +134,22 @@ let get_ocaml_error_message exn =
     Scanf.sscanf
       str
       "Characters %d-%d:\n%[\000-\255]"
-      (fun start stop msg -> ((start, stop), msg))
-  with _ ->
-    ((0, 0), str)
+      (fun start stop msg -> ((start, stop), msg, None))
+  with Scanf.Scan_failure(_) ->
+  try
+    Scanf.sscanf
+      str
+      "Line %d, characters %d-%d:\n%[\000-\255]"
+      (fun line start stop msg -> ((start, stop), msg, Some{start=line; stop=line}))
+  with Scanf.Scan_failure(_) ->
+  try
+    Scanf.sscanf
+      str
+      "Lines %d-%d, characters %d-%d:\n%[\000-\255]"
+      (fun start_line stop_line start stop msg -> ((start, stop),
+                                                   msg, Some{start=start_line;stop=stop_line}))
+  with Scanf.Scan_failure(_) ->
+    ((0, 0), str, None)
 
 let collect_formatters buf pps f =
   (* First flush all formatters. *)
@@ -176,10 +197,7 @@ let discard_formatters pps f =
   (* Output functions. *)
   let out_functions = {
     Format.out_string = (fun _ _ _ -> ()); out_flush = ignore;
-    out_newline = ignore; out_spaces = ignore
-#if OCAML_VERSION >= (4, 06, 0)
-      ; out_indent = ignore
-#endif
+    out_newline = ignore; out_spaces = ignore ; out_indent = ignore
   } in
   (* Replace formatter functions. *)
   List.iter (fun pp -> Format.pp_set_formatter_out_functions pp out_functions) pps;
@@ -242,14 +260,10 @@ let parse_default parse str eos_is_error =
         (* If the string is empty, do not report an error. *)
         raise Need_more
     | Lexer.Error (error, loc) ->
-#if OCAML_VERSION >= (4, 08, 0)
         (match Location.error_of_exn (Lexer.Error (error, loc)) with
         | Some (`Ok error)->
           Error ([mkloc loc], get_message Location.print_report error)
         | _-> raise Need_more)
-#else
-        Error ([mkloc loc], get_message Lexer.report_error error)
-#endif
     | Syntaxerr.Error error -> begin
       match error with
       | Syntaxerr.Unclosed (opening_loc, opening, closing_loc, closing) ->
@@ -273,10 +287,16 @@ let parse_default parse str eos_is_error =
       | Syntaxerr.Ill_formed_ast (loc, s) ->
           Error ([mkloc loc],
                  Printf.sprintf "Error: broken invariant in parsetree: %s" s)
-#if OCAML_VERSION >= (4, 03, 0)
       | Syntaxerr.Invalid_package_type (loc, s) ->
           Error ([mkloc loc],
                  Printf.sprintf "Invalid package type: %s" s)
+#if OCAML_VERSION >= (5, 0, 0)
+      | Syntaxerr.Removed_string_set loc ->
+          Error ([mkloc loc],
+            "Syntax error: strings are immutable, there is no assignment \
+             syntax for them.\n\
+             Hint: Mutable sequences of bytes are available in the Bytes module.\n\
+             Hint: Did you mean to use 'Bytes.set'?")
 #endif
     end
     | Syntaxerr.Escape_error | Parsing.Parse_error ->
@@ -309,12 +329,6 @@ let with_loc loc str = {
   Location.loc = loc;
 }
 
-#if OCAML_VERSION >= (4, 03, 0)
-let nolabel = Asttypes.Nolabel
-#else
-let nolabel = ""
-#endif
-
 (* Check that the given phrase can be evaluated without typing/compile
    errors. *)
 let check_phrase phrase =
@@ -335,13 +349,16 @@ let check_phrase phrase =
         let env = !Toploop.toplevel_env in
         (* Construct "let _ () = let module _ = struct <items> end in ()" in order to test
            the typing and compilation of [items] without evaluating them. *)
-        let unit = with_loc loc (Longident.Lident "()") in
+        let unit =
+          let (%.) a b = Longident.Ldot (a, b) in
+          with_loc loc (Lident "Stdlib" %. "Unit" %. "()")
+        in
         let top_def =
           let open Ast_helper in
           with_default_loc loc
             (fun () ->
                Str.eval
-                 (Exp.fun_ nolabel None (Pat.construct unit None)
+                 (Exp.fun_ Nolabel None (Pat.construct unit None)
                    (Exp.letmodule (with_loc loc
                         #if OCAML_VERSION >= (4, 10, 0)
                         (Some "_")
@@ -365,10 +382,10 @@ let check_phrase phrase =
           None
         with exn ->
           (* The phrase contains errors. *)
-          let loc, msg = get_ocaml_error_message exn in
+          let loc, msg, line = get_ocaml_error_message exn in
           Toploop.toplevel_env := env;
           Btype.backtrack snap;
-          Some ([loc], msg)
+          Some ([loc], msg, [line])
 
 (* +-----------------------------------------------------------------+
    | Prompt                                                          |
@@ -448,7 +465,7 @@ let make_prompt ui profile count size key_sequence (recording, macro_count, macr
               txta;
               Array.make
                 (size.cols - Array.length txta - Array.length txtb)
-                (Zed_char.unsafe_of_uChar (UChar.of_int 0x2500), { none with foreground = Some (color lcyan blue); bold = Some bold });
+                (Zed_char.of_utf8 "\u{2500}", { none with foreground = Some (color lcyan blue); bold = Some bold });
               txtb;
             ]
         ) second_line
@@ -469,38 +486,49 @@ let prompt = ref default_prompt
 
 let edit_mode= ref LTerm_editor.Default
 
+let default_info = {
+  Toploop.section = "UTop";
+  doc = ""; (* TODO: have some kind of documentation *)
+}
+
 let () =
-  Hashtbl.add Toploop.directive_table "utop_prompt_simple"
+  Toploop.add_directive "utop_prompt_simple"
     (Toploop.Directive_none
        (fun () ->
-         prompt := S.map (Printf.ksprintf LTerm_text.of_utf8 "utop [%d]: ") count));
+         prompt := S.map (Printf.ksprintf LTerm_text.of_utf8 "utop [%d]: ") count))
+    default_info;
 
-  Hashtbl.add Toploop.directive_table "utop_prompt_dummy"
+  Toploop.add_directive "utop_prompt_dummy"
     (Toploop.Directive_none
        (fun () ->
-         prompt := S.const (LTerm_text.of_utf8 "# ")));
+         prompt := S.const (LTerm_text.of_utf8 "# ")))
+    default_info;
 
-  Hashtbl.add Toploop.directive_table "utop_prompt_fancy_light"
+  Toploop.add_directive "utop_prompt_fancy_light"
     (Toploop.Directive_none
        (fun () ->
          set_profile Light;
-         prompt := default_prompt));
+         prompt := default_prompt))
+    default_info;
 
-  Hashtbl.add Toploop.directive_table "utop_prompt_fancy_dark"
+  Toploop.add_directive "utop_prompt_fancy_dark"
     (Toploop.Directive_none
        (fun () ->
          set_profile Dark;
-         prompt := default_prompt));
+         prompt := default_prompt))
+    default_info;
 
-  Hashtbl.add Toploop.directive_table "edit_mode_default"
+  Toploop.add_directive "edit_mode_default"
     (Toploop.Directive_none
        (fun () ->
-         edit_mode:= LTerm_editor.Default));
+         edit_mode:= LTerm_editor.Default))
+    default_info;
 
-  Hashtbl.add Toploop.directive_table "edit_mode_vi"
+  Toploop.add_directive "edit_mode_vi"
     (Toploop.Directive_none
        (fun () ->
          edit_mode:= LTerm_editor.Vi))
+    default_info
 
 (* +-----------------------------------------------------------------+
    | Help                                                            |
@@ -522,7 +550,7 @@ let doc_of_action action =
     LTerm_read_line.doc_of_action action
 
 let () =
-  Hashtbl.add Toploop.directive_table "utop_help"
+  Toploop.add_directive "utop_help"
     (Toploop.Directive_none
        (fun () ->
           print_endline "If you can't see the prompt properly try: #utop_prompt_simple
@@ -537,9 +565,10 @@ utop defines the following directives:
 #topfind_log     : display messages recorded from findlib since the beginning of the session
 #topfind_verbose : enable/disable topfind verbosity
 
-For a complete description of utop, look at the utop(1) manual page."));
+For a complete description of utop, look at the utop(1) manual page."))
+    default_info;
 
-  Hashtbl.add Toploop.directive_table "utop_bindings"
+  Toploop.add_directive "utop_bindings"
     (Toploop.Directive_none
        (fun () ->
           let make_lines keys actions acc =
@@ -608,9 +637,10 @@ For a complete description of utop, look at the utop(1) manual page."));
             output_string stdout (Buffer.contents buf)
           in
           List.iter format_line table;
-          flush stdout));
+          flush stdout))
+    default_info;
 
-  Hashtbl.add Toploop.directive_table "utop_macro"
+  Toploop.add_directive "utop_macro"
     (Toploop.Directive_none
        (fun () ->
           let macro = Zed_macro.contents LTerm_read_line.macro in
@@ -620,11 +650,13 @@ For a complete description of utop, look at the utop(1) manual page."));
                output_char stdout '\n')
             macro;
           flush stdout))
+    default_info
 
 let () =
-  Hashtbl.add Toploop.directive_table "pwd"
+  Toploop.add_directive "pwd"
     (Toploop.Directive_none
        (fun () -> print_endline (Sys.getcwd ())))
+    default_info
 
 let make_stash_directive entry_formatter fname =
   if get_ui () = Emacs then
@@ -660,7 +692,7 @@ let () =
     Printf.sprintf "(* %s *)" out
   end
   in
-  Hashtbl.add Toploop.directive_table "utop_stash" (Toploop.Directive_string fn)
+  Toploop.add_directive "utop_stash" (Toploop.Directive_string fn) default_info
 
 let () =
   let fn = make_stash_directive begin function
@@ -670,7 +702,7 @@ let () =
     out
   end
   in
-  Hashtbl.add Toploop.directive_table "utop_save" (Toploop.Directive_string fn)
+  Toploop.add_directive "utop_save" (Toploop.Directive_string fn) default_info
 
 (* +-----------------------------------------------------------------+
    | Findlib stuff                                                   |
@@ -702,19 +734,19 @@ let () =
     if S.value topfind_verbose then real_log str
 
 let () =
-  Hashtbl.add
-    Toploop.directive_table
+  Toploop.add_directive
     "topfind_log"
     (Toploop.Directive_none
        (fun () ->
          List.iter (fun str -> print_string str; print_char '\n')
            (S.value topfind_log);
-         flush  stdout));
+         flush  stdout))
+    default_info;
 
-  Hashtbl.add
-    Toploop.directive_table
+  Toploop.add_directive
     "topfind_verbose"
     (Toploop.Directive_bool set_topfind_verbose)
+    default_info
 
 let split_words str =
   let len = String.length str in
@@ -749,27 +781,19 @@ let require packages =
     handle_findlib_error exn
 
 let () =
-  Hashtbl.add
-    Toploop.directive_table
+  Toploop.add_directive
     "require"
     (Toploop.Directive_string
        (fun str -> require (split_words str)))
+    default_info
 
 (* +-----------------------------------------------------------------+
    | Backports                                                       |
    +-----------------------------------------------------------------+ *)
 
-
-let try_finally ~always work=
-#if OCAML_VERSION >= (4, 08, 0)
-    Misc.try_finally ~always work
-#else
-    Misc.try_finally work always
-#endif
-
 let use_output command =
   let fn = Filename.temp_file "ocaml" "_toploop.ml" in
-  try_finally ~always:(fun () ->
+  Misc.try_finally ~always:(fun () ->
     try Sys.remove fn with Sys_error _ -> ())
     (fun () ->
        match
@@ -784,11 +808,11 @@ let use_output command =
 
 let () =
   let name = "use_output" in
-  if not (Hashtbl.mem Toploop.directive_table name) then
-    Hashtbl.add
-      Toploop.directive_table
+  if UTop_compat.toploop_get_directive name = None then
+    Toploop.add_directive
       name
       (Toploop.Directive_string use_output)
+      default_info
 
 (* +-----------------------------------------------------------------+
    | Initialization                                                  |
@@ -809,13 +833,8 @@ let () =
    | Compiler-libs re-exports                                        |
    +-----------------------------------------------------------------+ *)
 
-#if OCAML_VERSION >= (4, 08, 0)
-let get_load_path ()= Load_path.get_paths ()
-let set_load_path path= Load_path.init path
-#else
-let get_load_path ()= !Config.load_path
-let set_load_path path= Config.load_path := path
-#endif
+let get_load_path () = Load_path.get_paths ()
+let set_load_path = UTop_compat.set_load_path
 
 (* +-----------------------------------------------------------------+
    | Deprecated                                                      |
