@@ -102,8 +102,7 @@ val set_margin_function : (LTerm_geom.size -> int option) -> unit
   *)
 
 val phrase_terminator : string signal
-  (** The phrase terminator. It is ";;" by default and ";" when you
-      use revised syntax. *)
+  (** The phrase terminator, ";;". *)
 
 val get_phrase_terminator : unit -> string
   (** Returns the value of {!phrase_terminator}. *)
@@ -254,6 +253,12 @@ type location = int * int
     (** Type of a string-location. It is composed of a start and stop
         offsets (in bytes). *)
 
+type lines = {
+  start: int;
+  stop:  int;
+}
+    (** Type for a range of lines in a buffer from start to stop. *)
+
 (** Result of a function processing a programx. *)
 type 'a result =
   | Value of 'a
@@ -310,12 +315,12 @@ val get_message : (Format.formatter -> 'a -> unit) -> 'a -> string
   (** [get_message printer x] applies [printer] on [x] and
       returns everything it prints as a string. *)
 
-val get_ocaml_error_message : exn -> location * string
+val get_ocaml_error_message : exn -> location * string * (lines option)
   (** [get_ocaml_error_message exn] returns the location and error
       message for the exception [exn] which must be an exception from
       the compiler. *)
 
-val check_phrase : Parsetree.toplevel_phrase -> (location list * string) option
+val check_phrase : Parsetree.toplevel_phrase -> (location list * string * lines option list) option
   (** [check_phrase phrase] checks that [phrase] can be executed
       without typing or compilation errors. It returns [None] if
       [phrase] is OK and an error message otherwise.
@@ -347,7 +352,16 @@ val set_load_path : string list -> unit
 (* These variables are not used and deprecated: *)
 
 val prompt_continue : LTerm_text.t React.signal ref
+[@@deprecated]
+
 val prompt_comment : LTerm_text.t React.signal ref
+[@@deprecated]
+
 val smart_accept : bool ref
+[@@deprecated]
+
 val new_prompt_hooks : (unit -> unit) LTerm_dlist.t
+[@@deprecated]
+
 val at_new_prompt : (unit -> unit) -> unit
+[@@deprecated]
