@@ -11,6 +11,7 @@
 
 open Types
 open LTerm_read_line
+open UTop_compat
 open UTop_token
 
 module String_set = Set.Make(String)
@@ -254,17 +255,18 @@ let parse_label tokens =
 
 let list_directives phrase_terminator =
   String_map.bindings
-    (Hashtbl.fold
-       (fun dir kind map ->
+    (List.fold_left
+       (fun map dir ->
           let suffix =
-            match kind with
-              | Toploop.Directive_none _ -> phrase_terminator
-              | Toploop.Directive_string _ -> " \""
-              | Toploop.Directive_bool _  | Toploop.Directive_int _ | Toploop.Directive_ident _ -> " "
+            match toploop_get_directive dir with
+              | Some (Toploop.Directive_none _) -> phrase_terminator
+              | Some (Toploop.Directive_string _) -> " \""
+              | Some (Toploop.Directive_bool _  | Toploop.Directive_int _ | Toploop.Directive_ident _) -> " "
+              | None -> assert false
           in
           String_map.add dir suffix map)
-       Toploop.directive_table
-       String_map.empty)
+       String_map.empty
+       (toploop_all_directive_names ()))
 
 (* +-----------------------------------------------------------------+
    | File listing                                                    |
@@ -392,11 +394,7 @@ let visible_modules () =
               (Sys.readdir (if dir = "" then Filename.current_dir_name else dir))
           with Sys_error _ ->
             acc)
-#if OCAML_VERSION >= (4, 08, 0)
         String_set.empty @@ Load_path.get_paths ()
-#else
-        String_set.empty !Config.load_path
-#endif
     )
 
 let field_name { ld_id = id } = Ident.name id
@@ -404,7 +402,7 @@ let constructor_name { cd_id = id } = Ident.name id
 
 let add_fields_of_type decl acc =
   match decl.type_kind with
-    | Type_variant constructors ->
+    | Type_variant _ ->
         acc
     | Type_record (fields, _) ->
         List.fold_left (fun acc field -> add (field_name field) acc) acc fields
@@ -415,7 +413,11 @@ let add_fields_of_type decl acc =
 
 let add_names_of_type decl acc =
   match decl.type_kind with
+#if OCAML_VERSION >= (4, 13, 0)
+    | Type_variant (constructors, _) ->
+#else
     | Type_variant constructors ->
+#endif
         List.fold_left (fun acc cstr -> add (constructor_name cstr) acc) acc constructors
     | Type_record (fields, _) ->
         List.fold_left (fun acc field -> add (field_name field) acc) acc fields
@@ -424,45 +426,22 @@ let add_names_of_type decl acc =
     | Type_open ->
         acc
 
-#if OCAML_VERSION >= (4, 08, 0)
 let path_of_mty_alias = function
   | Mty_alias path -> path
   | _ -> assert false
-#elif OCAML_VERSION >= (4, 04, 0)
-let path_of_mty_alias = function
-  | Mty_alias (_, path) -> path
-  | _ -> assert false
-#else
-let path_of_mty_alias = function
-  | Mty_alias path -> path
-  | _ -> assert false
-#endif
 
 let rec names_of_module_type = function
   | Mty_signature decls ->
       List.fold_left
         (fun acc decl -> match decl with
-#if OCAML_VERSION >= (4, 08, 0)
            | Sig_value (id, _, _)
            | Sig_typext (id, _, _, _)
            | Sig_module (id, _, _, _, _)
            | Sig_modtype (id, _, _)
            | Sig_class (id, _, _, _)
            | Sig_class_type (id, _, _, _) ->
-#else
-           | Sig_value (id, _)
-           | Sig_typext (id, _, _)
-           | Sig_module (id, _, _)
-           | Sig_modtype (id, _)
-           | Sig_class (id, _, _)
-           | Sig_class_type (id, _, _) ->
-#endif
                add (Ident.name id) acc
-#if OCAML_VERSION >= (4, 08, 0)
            | Sig_type (id, decl, _, _) ->
-#else
-           | Sig_type (id, decl, _) ->
-#endif
                add_names_of_type decl (add (Ident.name id) acc))
         String_set.empty decls
   | Mty_ident path -> begin
@@ -491,11 +470,7 @@ let rec fields_of_module_type = function
            | Sig_class _
            | Sig_class_type _ ->
                acc
-#if OCAML_VERSION >= (4, 08, 0)
            | Sig_type (_, decl, _, _) ->
-#else
-           | Sig_type (_, decl, _) ->
-#endif
                add_fields_of_type decl acc)
         String_set.empty decls
   | Mty_ident path -> begin
@@ -513,15 +488,6 @@ let rec fields_of_module_type = function
   | _ ->
       String_set.empty
 
-let lookup_module id env =
-#if OCAML_VERSION >= (4, 10, 0)
-  let path, decl = Env.find_module_by_name id env in
-  (path, decl.md_type)
-#else
-  let path = Env.lookup_module id env ~load:true in
-  (path, (Env.find_module path env).md_type)
-#endif
-
 let find_module path env = (Env.find_module path env).md_type
 
 let names_of_module longident =
@@ -565,11 +531,7 @@ let list_global_names () =
         loop (add_names_of_type decl (add (Ident.name id) acc)) summary
     | Env.Env_extension(summary, id, _) ->
         loop (add (Ident.name id) acc) summary
-#if OCAML_VERSION >= (4, 08, 0)
     | Env.Env_module(summary, id, _, _) ->
-#else
-    | Env.Env_module(summary, id, _) ->
-#endif
         loop (add (Ident.name id) acc) summary
     | Env.Env_modtype(summary, id, _) ->
         loop (add (Ident.name id) acc) summary
@@ -579,28 +541,18 @@ let list_global_names () =
         loop (add (Ident.name id) acc) summary
     | Env.Env_functor_arg(summary, id) ->
         loop (add (Ident.name id) acc) summary
-#if OCAML_VERSION >= (4, 08, 0)
     | Env.Env_persistent (summary, id) ->
         loop (add (Ident.name id) acc) summary
-#endif
-#if OCAML_VERSION >= (4, 04, 0)
     | Env.Env_constraints (summary, _) ->
         loop acc summary
-#endif
 #if OCAML_VERSION >= (4, 10, 0)
     | Env.Env_copy_types summary ->
         loop acc summary
-#elif OCAML_VERSION >= (4, 06, 0)
+#else
     | Env.Env_copy_types (summary, _) ->
         loop acc summary
 #endif
-#if OCAML_VERSION >= (4, 08, 0)
     | Env.Env_open(summary, path) ->
-#elif OCAML_VERSION >= (4, 07, 0)
-    | Env.Env_open(summary, _, path) ->
-#else
-    | Env.Env_open(summary, path) ->
-#endif
         match try Some (Path_map.find path !local_names_by_path) with Not_found -> None with
           | Some names ->
               loop (String_set.union acc names) summary
@@ -640,11 +592,7 @@ let list_global_fields () =
         loop (add_fields_of_type decl (add (Ident.name id) acc)) summary
     | Env.Env_extension(summary, id, _) ->
         loop (add (Ident.name id) acc) summary
-#if OCAML_VERSION >= (4, 08, 0)
     | Env.Env_module(summary, id, _, _) ->
-#else
-    | Env.Env_module(summary, id, _) ->
-#endif
         loop (add (Ident.name id) acc) summary
     | Env.Env_functor_arg(summary, id) ->
         loop (add (Ident.name id) acc) summary
@@ -654,30 +602,18 @@ let list_global_fields () =
         loop (add (Ident.name id) acc) summary
     | Env.Env_cltype(summary, id, _) ->
         loop (add (Ident.name id) acc) summary
-#if OCAML_VERSION >= (4, 08, 0)
     | Env.Env_persistent (summary, id) ->
         loop (add (Ident.name id) acc) summary
-#endif
-#if OCAML_VERSION >= (4, 04, 0)
     | Env.Env_constraints (summary, _) ->
         loop acc summary
-#endif
 #if OCAML_VERSION >= (4, 10, 0)
     | Env.Env_copy_types summary ->
         loop acc summary
-#elif OCAML_VERSION >= (4, 06, 0)
+#else
     | Env.Env_copy_types (summary, _) ->
         loop acc summary
 #endif
-#if OCAML_VERSION >= (4, 07, 0)
-  #if OCAML_VERSION >= (4, 08, 0)
-      | Env.Env_open(summary, path) ->
-  #else
-      | Env.Env_open(summary, _, path) ->
-  #endif
-#else
     | Env.Env_open(summary, path) ->
-#endif
         match try Some (Path_map.find path !local_fields_by_path) with Not_found -> None with
           | Some fields ->
               loop (String_set.union acc fields) summary
@@ -703,7 +639,7 @@ let global_fields () = get_cached global_fields list_global_fields
    +-----------------------------------------------------------------+ *)
 
 let rec find_method meth type_expr =
-  match type_expr.desc with
+  match get_desc type_expr with
     | Tlink type_expr ->
         find_method meth type_expr
     | Tobject (type_expr, _) ->
@@ -727,7 +663,7 @@ let rec find_method meth type_expr =
         None
 
 let rec methods_of_type acc type_expr =
-  match type_expr.desc with
+  match get_desc type_expr with
     | Tlink type_expr ->
         methods_of_type acc type_expr
     | Tobject (type_expr, _) ->
@@ -759,13 +695,6 @@ let rec find_object meths type_expr =
               None
 
 let methods_of_object longident meths =
-  let lookup_value=
-#if OCAML_VERSION >= (4, 10, 0)
-    Env.find_value_by_name
-#else
-    Env.lookup_value
-#endif
-  in
   match lookup_env lookup_value longident !Toploop.toplevel_env with
     | None ->
         []
@@ -781,20 +710,12 @@ let methods_of_object longident meths =
    +-----------------------------------------------------------------+ *)
 
 let rec labels_of_type acc type_expr =
-  match type_expr.desc with
+  match get_desc type_expr with
     | Tlink te ->
         labels_of_type acc te
     | Tpoly (te, _) ->
         labels_of_type acc te
     | Tarrow(label, _, te, _) ->
-#if OCAML_VERSION < (4, 03, 0)
-        if label = "" then
-          labels_of_type acc te
-        else if label.[0] = '?' then
-          labels_of_type (String_map.add (String.sub label 1 (String.length label - 1)) Optional acc) te
-        else
-          labels_of_type (String_map.add label Required acc) te
-#else
       (match label with
       | Nolabel ->
         labels_of_type acc te
@@ -802,7 +723,6 @@ let rec labels_of_type acc type_expr =
         labels_of_type (String_map.add label Optional acc) te
       | Labelled label ->
         labels_of_type (String_map.add label Required acc) te)
-#endif
     | Tconstr(path, _, _) -> begin
         match lookup_env Env.find_type path !Toploop.toplevel_env with
           | None
@@ -815,13 +735,6 @@ let rec labels_of_type acc type_expr =
         String_map.bindings acc
 
 let labels_of_function longident meths =
-  let lookup_value=
-#if OCAML_VERSION >= (4, 10, 0)
-    Env.find_value_by_name
-#else
-    Env.lookup_value
-#endif
-  in
   match lookup_env lookup_value longident !Toploop.toplevel_env with
     | None ->
         []
@@ -833,13 +746,6 @@ let labels_of_function longident meths =
               labels_of_type String_map.empty type_expr
 
 let labels_of_newclass longident =
-  let lookup_class=
-#if OCAML_VERSION >= (4, 10, 0)
-    Env.find_class_by_name
-#else
-    Env.lookup_class
-#endif
-  in
   match lookup_env lookup_class longident !Toploop.toplevel_env with
     | None ->
         []
@@ -892,14 +798,6 @@ and find_context_in_quotation = function
    | Completion                                                      |
    +-----------------------------------------------------------------+ *)
 
-#if OCAML_VERSION < (4, 11, 0)
-let longident_parse= Longident.parse
-#else
-let longident_parse str=
-  let lexbuf= Lexing.from_string str in
-  Parse.longident lexbuf
-#endif
-
 let complete ~phrase_terminator ~input =
   let true_name, false_name = ("true", "false") in
   let tokens = UTop_lexer.lex_string input in
@@ -955,11 +853,7 @@ let complete ~phrase_terminator ~input =
               (fun acc d -> add_files filter acc (Filename.concat d dir))
               String_map.empty
               (Filename.current_dir_name ::
-#if OCAML_VERSION >= (4, 08, 0)
                 (Load_path.get_paths ())
-#else
-                !Config.load_path
-#endif
               )
 
           else
@@ -1019,11 +913,7 @@ let complete ~phrase_terminator ~input =
               (fun acc d -> add_files filter acc (Filename.concat d dir))
               String_map.empty
               (Filename.current_dir_name ::
-#if OCAML_VERSION >= (4, 08, 0)
                 (Load_path.get_paths ())
-#else
-                !Config.load_path
-#endif
               )
           else
             add_files filter String_map.empty (Filename.dirname file)
@@ -1045,7 +935,7 @@ let complete ~phrase_terminator ~input =
     (* Generic completion on directives. *)
     | [(Symbol "#", _); ((Lident dir | Uident dir), _); (Blanks, { idx2 = stop })] ->
         (stop,
-         match try Some (Hashtbl.find Toploop.directive_table dir) with Not_found -> None with
+         match toploop_get_directive dir with
            | Some (Toploop.Directive_none _) -> [(phrase_terminator, "")]
            | Some (Toploop.Directive_string _) -> [(" \"", "")]
            | Some (Toploop.Directive_bool _) -> [(true_name, phrase_terminator); (false_name, phrase_terminator)]
@@ -1053,7 +943,7 @@ let complete ~phrase_terminator ~input =
            | Some (Toploop.Directive_ident _) -> List.map (fun w -> (w, "")) (String_set.elements (global_names ()))
            | None -> [])
     | (Symbol "#", _) :: ((Lident dir | Uident dir), _) :: tokens -> begin
-        match try Some (Hashtbl.find Toploop.directive_table dir) with Not_found -> None with
+        match toploop_get_directive dir with
           | Some (Toploop.Directive_none _) ->
               (0, [])
           | Some (Toploop.Directive_string _) ->
